/*
 * Copyright 2006-2007, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log.h"
#include "network.h"
#include "sysstat.h"
#include "terminal.h"


int Fslines = 5, Showzfs = 1, Localmode = 0;
unsigned Interval = 5;
extern int Port;
extern char *McastAddr;

static sysstat_t
	*SS = 0, *Current = 0;
static unsigned
	Curses = 1;

static void
dealloc(sysstat_t *s)
{
	userstat_t *us;
	fsstat_t *fs;
	if (s == 0)
		return;
	us = s->us;
	while (us) {
		userstat_t *n = us->next;
		free(us->username);
		free(us);
		us = n;
	}
	fs = s->fs;
	while (fs) {
		fsstat_t *n = fs->next;
		free(fs->dev);
		free(fs->mnt);
		free(fs);
		fs = n;
	}
	free(s);
}


static void
add_stats(sysstat_t *s)
{
	sysstat_t *i = SS;
	assert((s != i) && (s->next == 0) && (s->prev == 0));
	/* search, if we already have an entry for this host */
	while (i && s) {
		assert(i->next && i->prev);
		if (strcmp(i->Hostinfo,s->Hostinfo)) {
			i = i->next;
			if (i == SS)
				break;
			continue;
		}
		if (i->TS) {
			/* replace this entry */
			if (i->prev != i) {
				assert(i->next != i);
				s->prev = i->prev;
				assert(s->prev->next == i);
				s->prev->next = s;
				s->next = i->next;
				assert(s->next->prev == i);
				s->next->prev = s;
			} else {
				assert(i->next == i);
				s->prev = s;
				s->next = s;
			}
			if (Current == i)
				Current = s;
			if (SS == i)
				SS = s;
			dealloc(i);
			assert(s->prev && s->next);
			return;
		} else {
			/* add info to this entry */
			/* preserve existing values */
			userstat_t *us = i->us;
			fsstat_t *fs = i->fs;
			sysstat_t *next = i->next, *prev = i->prev;
			memcpy(i,s,sizeof(sysstat_t));
			i->next = next;
			i->prev = prev;
			if (i->fs) {
				fsstat_t *lf = i->fs;
				while (lf->next)
					lf = lf->next;
				lf->next = fs;
			} else
				i->fs = fs;
			if (i->us) {
				userstat_t *lu =i->us;
				while (lu->next)
					lu = lu->next;
				lu->next = us;
			} else
				i->us = us;
			free(s);
			return;
		}
	}
	/* new sysstat not already in our list */
	assert(s);
	if (SS) {
		s->next = SS;
		s->prev = SS->prev;
		assert(s->prev->next == SS);
		s->prev->next = s;
		SS->prev = s;
	} else {
		s->next = s;
		s->prev = s;
		SS = s;
		Current = s;
	}
	assert(s->prev && s->next);
}


static void
process_input(int s, int itv)
{
	struct pollfd fds[2];
	char c;
	sysstat_t *st;
	
	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	fds[1].fd = s;
	fds[1].events = POLLIN;
	if (-1 == poll(fds,s == -1 ? 1 : sizeof(fds)/sizeof(struct pollfd),itv * 1000)) {
		if (errno != EINTR)
			error("error polling: %s\n",strerror(errno));
		return;
	}
	if (1 == read(STDIN_FILENO,&c,sizeof(char))) {
		switch (c) {
		case 'h':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->next;
			break;
		case 'H':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->prev;
			break;
		case 'n':
			set_sort_by_nproc();
			if (Current)
				++Current->TS;
			break;
		case 'q':
			exit(EXIT_SUCCESS);
			break;
		case 'r':
			set_sort_by_rss();
			if (Current)
				++Current->TS;
			break;
		case 's':
			set_sort_by_size();
			if (Current)
				++Current->TS;
			break;
		case 't':
			set_sort_by_time();
			if (Current)
				++Current->TS;
			break;
		case 'u':
			set_sort_by_user();
			if (Current)
				++Current->TS;
			break;
		case 'c':
			set_sort_by_pcpu();
			if (Current)
				++Current->TS;
			break;
		case 'z':
			Showzfs ^= 1;
			if (Current)
				++Current->TS;
			break;
		case '+':
			set_sort_by_hs();
			if (Current)
				++Current->TS;
			break;
		case '?':
			if (Curses) {
				display_help();
				if (Current)
					++Current->TS;
			}
			break;
		default:
			;
		}
	}
	if ((s != -1) && (st = receive_stats(s))) {
		assert(st != SS);
		add_stats(st);
	}
}


static void
usage(int r)
{
	printf(
	"sysstat [option] [interval [count]]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-c        : write stats to stdout instead of using curses\n"
	"-d        : start sysstat as daemon, multicasting stats\n"
	"-f <num>  : show at least num lines of filesystem information\n"
	"-k        : kill sysstat's daemon\n"
	"-l        : start sysstat in local-mode (i.e. without network support)\n"
	"-m <addr> : send info to spawndaemon address <addr> (default: 224.0.0.75)\n"
	"-p <port> : set spawndaemon port to <port> (default: 7536)\n"
	"-r        : sort user/process summary by rss\n"
	"-s        : sort user/process summary by size\n"
	"-t        : sort user/process summary by %%CPU time\n"
	"-u        : sort user/process summary by username\n"
	"-v        : operate in verbose mode\n"
	"-V        : display version number and exit\n"
	);
	exit(r);
}


static void
version(void)
{
	printf(
	"sysstat version "VERSION"\n"
	"Copyright 2006-2007, Thomas Maier-Komor\n"
	"This software is licensed under the GPLv2. See file COPYING for details.\n"
	);
	exit(EXIT_SUCCESS);
}


int
main(int argc, char **argv)
{
	unsigned daemon = 0, spawndaemon = 1,
		 iter = 0, killdaemon = 0;
	int c, count = -1, pid, sock;
	char *mcastAddr = 0, *port = 0, *itv = 0, *cnt = 0;
	extern char *optarg;
	FILE *pidf;

	while ((c = getopt(argc, argv, "cdf:hklm:p:rstuvV")) != -1) {
		int tmp;
		switch(c) {
		default:
			assert(0);
			break;
		case 'c':
			Curses = 0;
			break;
		case 'd':
			daemon = 1;
			break;
		case 'f':
			if (1 != sscanf(optarg,"%u",&Fslines))
				usage(EXIT_FAILURE);
			break;
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case 'k':
			killdaemon = 1;
			break;
		case 'l':
			Localmode = 1;
			break;
		case 'm':
			McastAddr = optarg;
			mcastAddr = argv[optind-1];
			info("spawndaemon address set to %s\n",optarg);
			break;
		case 'p':
			if (1 != sscanf(optarg,"%u",&Port))
				usage(EXIT_FAILURE);
			info("port set to %s\n",Port);
			port = argv[optind-1];
			break;
		case 'r':
			set_sort_by_rss();
			break;
		case 's':
			set_sort_by_size();
			break;
		case 't':
			set_sort_by_time();
			break;
		case 'u':
			set_sort_by_user();
			break;
		case 'v':
			setVerbose("sysstat.log");
			break;
		case 'V':
			version();
			break;
		case '?':
			usage(EXIT_FAILURE);
			break;
		}
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%u",&Interval))
			usage(EXIT_FAILURE);
		dbug("interval set to %u\n",Interval);
		++optind;
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%d",&count) || (count <= 0))
			usage(EXIT_FAILURE);
		++optind;
	}
	if (optind < argc) {
		usage(EXIT_FAILURE);
	}
	if (-1 == fcntl(STDIN_FILENO,F_SETFL, O_NONBLOCK))
		error("Cannot set input file to non-blocking: %s\n",strerror(errno));
	pidf = fopen("/tmp/sysstat.pid","r");
	if ((Localmode == 0) && (pidf != 0)) {
		char pn[] = "/proc/\0\0\0\0\0\0",*p;
		struct stat st;
		info("/tmp/sysstat.pid exists\n");
		if (1 != fscanf(pidf,"%d",&pid)) {
			warn("corrupt pid file /tmp/systat.pid - unlinking it\n");
			if (-1 == unlink("/tmp/sysstat.pid"))
				warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
		} else {
			p = pn + 6;
			sprintf(p,"%d",pid);
			info("stating %s\n",pn);
			if (-1 == stat(pn,&st)) {
				info("unlinking stale pid file /tmp/sysstat.pid\n");
				if (-1 == unlink("/tmp/sysstat.pid"))
					warn("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
			} else {
				info("daemon already running with pid %d\n",pid);
				spawndaemon = 0;
			}
		}
		fclose(pidf);
	}
	if (killdaemon) {
		if (pid == -1) {
			error("cannot kill sysstat daemon: no sysstat daemon running\n");
			exit(EXIT_FAILURE);
		}
		if (0 == kill(pid,SIGTERM)) {
			info("sending SIGTERM to %d\n",pid);
			exit(EXIT_SUCCESS);
		} else {
			error("cannot send SIGTERM to %d: %s\n",pid,strerror(errno));
			exit(EXIT_FAILURE);
		}
	}
	if ((Localmode == 0) && (spawndaemon != 0)) {
		pid_t pid, dpid;
		if ((pid = vfork())) {
			int status;
			waitpid(pid, &status, 0);
			assert(status == 0);
		} else if (dpid = vfork()) {
			exit(EXIT_SUCCESS);
		} else {
			int i;
			char bin[PATH_MAX];
			char *args[16];
			int num = 1;
			strcpy(bin,getexecname());
			dirname(bin);
			strcat(bin,"/../sbin/sysstatd");
			args[0] = bin;
			args[1] = "-v";
			++num;
			if (port)
				args[num++] = port;
			if (mcastAddr)
				args[num++] = mcastAddr;
			if (itv)
				args[num++] = itv;
			if (cnt)
				args[num++] = cnt;
			args[num] = 0;
			dbug("starting sysstatd binary %s\n",bin);
			for (i = 0; args[i]; ++i)
				dbug("\targ %d: %s\n",i,args[i]);
			execv(bin,args);
			error("could not exec sysstatd: %s\n",strerror(errno));
		} 
		if (daemon)
			exit(EXIT_SUCCESS);
	} else if (Localmode == 0)
		printf("sysstat daemon already running: waiting for data...\n");

	sock = setup_source();
	if (Curses)
		Curses = init_curses();
	while ((count == 0) || (iter < count)) {
		if (Curses) {
			iter += display_curses(Current);
		} else {
			iter += display_stdout(Current);
		}
		process_input(sock,-1);
	}
	return 0;
}

// vim:tw=0
